/* Generated by JASS Version 0.1 64-bit DEBUGGING-build Copyright (c) 2016-2017 Andrew Trotman, University of Otago */
#include <stdint.h>

#include"JASS_postings.h"
#include"JASS_vocabulary.h"
JASS_ci_vocab dictionary[] = {
{"6",T_6},
{"1",T_1},
{"4",T_4},
{"5",T_5},
{"3",T_3},
{"8",T_8},
{"7",T_7},
{"2",T_2},
{"9",T_9},
{"10",T_10},
{"four",T_four},
{"eight",T_eight},
{"five",T_five},
{"seven",T_seven},
{"two",T_two},
{"six",T_six},
{"three",T_three},
{"one",T_one},
{"nine",T_nine},
{"ten",T_ten},
};
uint64_t dictionary_length = 20;
