/* Generated by JASS Version 0.1 64-bit DEBUGGING-build Copyright (c) 2016-2017 Andrew Trotman, University of Otago */
#include <string>
#include <vector>
std::vector<std::string> primary_key =
{
"-",
"1",
"2",
"3",
"4",
"5",
"6",
"7",
"8",
"9",
"10",
};
