/*
	COVERITY_MODEL.C
	----------------
	Copyright (c) 2016 Andrew Trotman
	Released under the 2-clause BSD license (See:https://en.wikipedia.org/wiki/BSD_licenses)
*/
/*!
	@file
	@brief List of exceptions for Coverity Scan
	@author Andrew Trotman
	@copyright 2016 Andrew Trotman
*/
