/* Generated by JASS Version 0.1 64-bit DEBUGGING-build Copyright (c) 2016-2017 Andrew Trotman, University of Otago */
#include <stddef.h>
#include <stdint.h>
#include"query_heap.h"

using namespace JASS;
void T_6(query_heap &q)
{
q.add_rsv(6,1);
}
void T_1(query_heap &q)
{
q.add_rsv(1,1);
}
void T_4(query_heap &q)
{
q.add_rsv(4,1);
}
void T_5(query_heap &q)
{
q.add_rsv(5,1);
}
void T_3(query_heap &q)
{
q.add_rsv(3,1);
}
void T_8(query_heap &q)
{
q.add_rsv(8,1);
}
void T_7(query_heap &q)
{
q.add_rsv(7,1);
}
void T_2(query_heap &q)
{
q.add_rsv(2,1);
}
void T_9(query_heap &q)
{
q.add_rsv(9,1);
}
void T_10(query_heap &q)
{
q.add_rsv(10,1);
}
void T_four(query_heap &q)
{
q.add_rsv(7,1);
q.add_rsv(8,1);
q.add_rsv(9,1);
q.add_rsv(10,1);
}
void T_eight(query_heap &q)
{
q.add_rsv(3,1);
q.add_rsv(4,1);
q.add_rsv(5,1);
q.add_rsv(6,1);
q.add_rsv(7,1);
q.add_rsv(8,1);
q.add_rsv(9,1);
q.add_rsv(10,1);
}
void T_five(query_heap &q)
{
q.add_rsv(6,1);
q.add_rsv(7,1);
q.add_rsv(8,1);
q.add_rsv(9,1);
q.add_rsv(10,1);
}
void T_seven(query_heap &q)
{
q.add_rsv(4,1);
q.add_rsv(5,1);
q.add_rsv(6,1);
q.add_rsv(7,1);
q.add_rsv(8,1);
q.add_rsv(9,1);
q.add_rsv(10,1);
}
void T_two(query_heap &q)
{
q.add_rsv(9,1);
q.add_rsv(10,1);
}
void T_six(query_heap &q)
{
q.add_rsv(5,1);
q.add_rsv(6,1);
q.add_rsv(7,1);
q.add_rsv(8,1);
q.add_rsv(9,1);
q.add_rsv(10,1);
}
void T_three(query_heap &q)
{
q.add_rsv(8,1);
q.add_rsv(9,1);
q.add_rsv(10,1);
}
void T_one(query_heap &q)
{
q.add_rsv(10,1);
}
void T_nine(query_heap &q)
{
q.add_rsv(2,1);
q.add_rsv(3,1);
q.add_rsv(4,1);
q.add_rsv(5,1);
q.add_rsv(6,1);
q.add_rsv(7,1);
q.add_rsv(8,1);
q.add_rsv(9,1);
q.add_rsv(10,1);
}
void T_ten(query_heap &q)
{
q.add_rsv(1,1);
q.add_rsv(2,1);
q.add_rsv(3,1);
q.add_rsv(4,1);
q.add_rsv(5,1);
q.add_rsv(6,1);
q.add_rsv(7,1);
q.add_rsv(8,1);
q.add_rsv(9,1);
q.add_rsv(10,1);
}
